@@ -119,6 +119,7 @@ typedef enum {
 	opt_use_ldaps,
 	opt_account_disable,
 	opt_ldap_passwd,
+	opt_recursive_delete,
 } Option;
 
 static adcli_tool_desc common_usages[] = {
@@ -171,6 +172,7 @@ static adcli_tool_desc common_usages[] = {
 	                      "to the Samba specific configuration database" },
 	{ opt_samba_data_tool, "Absolute path to the tool used for add-samba-data" },
 	{ opt_ldap_passwd, "Use LDAP add/mod operation to set/change password" },
+	{ opt_recursive_delete, "Delete computer object and child objects" },
 	{ opt_verbose, "show verbose progress and failure messages", },
 	{ 0 },
 };
@@ -363,6 +365,7 @@ parse_option (Option opt,
 	case opt_one_time_password:
 	case opt_add_samba_data:
 	case opt_ldap_passwd:
+	case opt_recursive_delete:
 		assert (0 && "not reached");
 		break;
 	}
@@ -517,6 +520,8 @@ adcli_tool_computer_join (adcli_conn *conn,
 	else if (show_password)
 		dump_password (conn, enroll);
 
+	ensure_host_keytab_selinux_context (ADCLI_SUCCESS, enroll);
+
 	adcli_enroll_unref (enroll);
 
 	return 0;
@@ -650,6 +655,8 @@ adcli_tool_computer_update (adcli_conn *conn,
 	else if (show_password)
 		dump_password (conn, enroll);
 
+	ensure_host_keytab_selinux_context (ADCLI_SUCCESS, enroll);
+
 	adcli_enroll_unref (enroll);
 
 	return 0;
@@ -718,6 +725,13 @@ adcli_tool_computer_testjoin (adcli_conn *conn,
 		return -res;
 	}
 
+	/* Use realm from keytab as domain name if not set explicitly because
+	 * this is most probably the name used during join and the DNS
+	 * hostname might be different. */
+	if (adcli_conn_get_domain_name (conn) == NULL) {
+		adcli_conn_set_domain_name (conn, adcli_conn_get_domain_realm (conn));
+	}
+
 	res = adcli_conn_connect (conn);
 	if (res != ADCLI_SUCCESS) {
 		adcli_enroll_unref (enroll);
@@ -727,7 +741,7 @@ adcli_tool_computer_testjoin (adcli_conn *conn,
 		return -res;
 	}
 
-	printf ("Sucessfully validated join to domain %s\n",
+	printf ("Successfully validated join to domain %s\n",
 	        adcli_conn_get_domain_name (conn));
 
 	adcli_enroll_unref (enroll);
@@ -948,12 +962,14 @@ adcli_tool_computer_delete (adcli_conn *conn,
 	adcli_enroll *enroll;
 	adcli_result res;
 	int opt;
+	adcli_enroll_flags flags = 0;
 
 	struct option options[] = {
 		{ "domain", required_argument, NULL, opt_domain },
 		{ "domain-realm", required_argument, NULL, opt_domain_realm },
 		{ "domain-controller", required_argument, NULL, opt_domain_controller },
 		{ "use-ldaps", no_argument, 0, opt_use_ldaps },
+		{ "recursive", no_argument, 0, opt_recursive_delete },
 		{ "login-user", required_argument, NULL, opt_login_user },
 		{ "login-ccache", optional_argument, NULL, opt_login_ccache },
 		{ "no-password", no_argument, 0, opt_no_password },
@@ -977,6 +993,9 @@ adcli_tool_computer_delete (adcli_conn *conn,
 
 	while ((opt = adcli_tool_getopt (argc, argv, options)) != -1) {
 		switch (opt) {
+		case opt_recursive_delete:
+			flags |= ADCLI_ENROLL_RECURSIVE_DELETE;
+			break;
 		case 'h':
 		case '?':
 		case ':':
@@ -1023,7 +1042,7 @@ adcli_tool_computer_delete (adcli_conn *conn,
 	if (argc == 1)
 		parse_fqdn_or_name (enroll, argv[0]);
 
-	res = adcli_enroll_delete (enroll, 0);
+	res = adcli_enroll_delete (enroll, flags);
 	if (res != ADCLI_SUCCESS) {
 		warnx ("deleting %s in %s domain failed: %s", argv[0],
 		       adcli_conn_get_domain_name (conn),
@@ -1256,6 +1275,8 @@ adcli_tool_computer_managed_service_account (adcli_conn *conn,
 	else if (show_password)
 		dump_password (conn, enroll);
 
+	ensure_host_keytab_selinux_context (ADCLI_SUCCESS, enroll);
+
 	adcli_enroll_unref (enroll);
 
 	return 0;
