@@ -36,6 +36,7 @@
 #include <paths.h>
 #include <stdio.h>
 #include <unistd.h>
+#include <termios.h>
 
 
 static char *adcli_temp_directory = NULL;
@@ -63,6 +64,7 @@ struct {
 	{ "create-msa", adcli_tool_computer_managed_service_account, "Create a managed service account in the given AD domain", },
 	{ "create-user", adcli_tool_user_create, "Create a user account", },
 	{ "delete-user", adcli_tool_user_delete, "Delete a user account", },
+	{ "passwd-user", adcli_tool_user_passwd, "(Re)set a user password", },
 	{ "create-group", adcli_tool_group_create, "Create a group", },
 	{ "delete-group", adcli_tool_group_delete, "Delete a group", },
 	{ "add-member", adcli_tool_member_add, "Add users to a group", },
@@ -73,6 +75,11 @@ struct {
 static char
 short_option (int opt)
 {
+	// isalpha and isdigit require the argument to be representable as unsigned char
+	// if they're not representable as unsigned char, then we assume they're not printable -
+	// matching for example the enumerator definitions in computer.c
+	if (opt != (int)(unsigned char)opt)
+		return 0;
 	if (isalpha (opt) || isdigit (opt))
 		return (char)opt;
 	return 0;
@@ -202,6 +209,47 @@ command_usage (void)
 	printf ("\nSee 'adcli <command> --help' for more information\n");
 }
 
+static char *get_password (const char *prompt)
+{
+	int ret;
+	struct termios termios;
+	struct termios orig_termios;
+	char *buf = NULL;
+	size_t buf_len = 0;
+
+	ret = tcgetattr (fileno (stdin), &termios);
+	if (ret != 0 ) {
+		return NULL;
+	}
+
+	orig_termios = termios;
+	termios.c_lflag &= ~ECHO;
+
+	ret = tcsetattr (fileno (stdin), TCSAFLUSH, &termios);
+	if (ret != 0) {
+		return NULL;
+	}
+
+	fprintf (stdout, "%s", prompt);
+	fflush (stdout);
+
+	ret = getline (&buf, &buf_len, stdin);
+	tcsetattr (fileno (stdin), TCSAFLUSH, &orig_termios);
+	if (ret <= 0) {
+		free (buf);
+		return NULL;
+	}
+
+	if (buf[ret - 1] == '\n') {
+		/* remove new-line character from the end of the buffer and
+		 * echo it to stdout */
+		buf[ret - 1] = '\0';
+		fprintf (stdout, "\n");
+	}
+
+	return buf;
+}
+
 char *
 adcli_prompt_password_func (adcli_login_type login_type,
                             const char *name,
@@ -215,7 +263,7 @@ adcli_prompt_password_func (adcli_login_type login_type,
 	if (asprintf (&prompt, "Password for %s: ", name) < 0)
 		return_val_if_reached (NULL);
 
-	password = getpass (prompt);
+	password = get_password (prompt);
 	free (prompt);
 
 	if (password == NULL)
@@ -223,6 +271,7 @@ adcli_prompt_password_func (adcli_login_type login_type,
 
 	result = strdup (password);
 	adcli_mem_clear (password, strlen (password));
+	free (password);
 
 	return result;
 }
@@ -337,20 +386,37 @@ setup_krb5_conf_directory (adcli_conn *conn)
 	}
 
 	if (!failed) {
-		if (mkdtemp (directory) == NULL) {
+		mode_t old_umask;
+		char *dtemp = NULL;
+
+		old_umask = umask (0077);
+		dtemp = mkdtemp (directory);
+		umask (old_umask);
+
+		if (dtemp == NULL) {
 			errn = errno;
 			failed = 1;
 			warnx ("couldn't create temporary directory in: %s: %s",
 			       parent, strerror (errn));
 		} else {
-			if (asprintf (&filename, "%s/krb5.conf", directory) < 0 ||
-			    asprintf (&snippets, "%s/krb5.d", directory) < 0 ||
-			    asprintf (&contents, "includedir %s\n%s%s\n", snippets,
-			              krb5_conf ? "include " : "",
-			              krb5_conf ? krb5_conf : "") < 0) {
+			if (asprintf (&filename, "%s/krb5.conf", directory) < 0) {
+				warnx ("unexpected: out of memory");
+				failed = 1;
+			}
+			if (!failed && asprintf (&snippets, "%s/krb5.d", directory) < 0) {
+				free (filename);
+				filename = NULL;
+				warnx ("unexpected: out of memory");
+				failed = 1;
+			}
+			if (!failed && asprintf (&contents, "includedir %s\n%s%s\n", snippets,
+			                         krb5_conf ? "include " : "",
+			                         krb5_conf ? krb5_conf : "") < 0) {
+				free (snippets);
+				snippets = NULL;
+				free (filename);
+				filename = NULL;
 				warnx ("unexpected: out of memory");
-				filename = NULL; /* content is undefined */
-				snippets = NULL; /* content is undefined */
 				contents = NULL; /* content is undefined */
 				failed = 1;
 			}
@@ -538,6 +604,12 @@ main (int argc,
 
 		if (conn)
 			adcli_conn_unref (conn);
+#ifdef VENDOR_MSG
+		if (ret != 0) {
+			fprintf (stderr, VENDOR_MSG"\n");
+		}
+#endif
+
 		return ret;
 	}
 
